"""SQL table class"""

from abc import ABC, abstractmethod
from typing import Any, Collection, Iterable, TYPE_CHECKING, Optional, Sequence
from sqlcontroller.querybuilder import SqliteQueryBuilder
from sqlcontroller.field import Field

if TYPE_CHECKING:
    from sqlcontroller.controller import AbstractSqlController, SqliteController


class DbTable(ABC):
    """Provide base for database table classes"""

    controller: "AbstractSqlController"
    name: str

    def _execute(self, query: str, values: Optional[Collection] = None) -> Any:
        """Forward execution to controller"""
        self.controller.execute(query, self.name, values)

    def _executemany(self, query: str, valuelists: Optional[Collection] = None) -> Any:
        """Forward execution to controller"""
        self.controller.executemany(query, self.name, valuelists)

    @abstractmethod
    def add_row(
        self, values: Collection, fields: Optional[Iterable[Field]] = None
    ) -> None:
        """Add new row to a table"""

    @abstractmethod
    def add_rows(
        self, valuelists: Sequence[Collection], fields: Optional[Iterable[Field]] = None
    ) -> None:
        """Add new row to a table"""

    @abstractmethod
    def count_rows(self, clause: str = "") -> int:
        """Count rows"""

    @abstractmethod
    def get_row(
        self, fields: Iterable = None, clause: str = "", as_dict: bool = False
    ) -> Any:
        """Get first matching row from a table"""

    @abstractmethod
    def get_rows(
        self, fields: Iterable = None, clause: str = "", as_dicts: bool = False
    ) -> list[Any]:
        """Get all matching rows from a table"""

    @abstractmethod
    def update_rows(self, values: dict, clause: str) -> None:
        """Modify a table's row's values"""

    @abstractmethod
    def delete_rows(self, clause: str) -> None:
        """Remove matching rows from a table"""

    @abstractmethod
    def delete_all_rows(self) -> None:
        """Remove all rows from a table"""


IterFieldOpt = Optional[Iterable[Field]]


class SqliteTable(DbTable):
    """Define methods to operate on database table"""

    controller: "SqliteController"

    def __init__(self, name: str, controller: "SqliteController") -> None:
        self.name = name
        self.controller = controller

    def add_row(self, values: Collection, fields: IterFieldOpt = None) -> None:
        query = SqliteQueryBuilder.build_insert_query(values, fields)
        self._execute(query, values)

    def add_rows(
        self, valuelists: Sequence[Collection], fields: IterFieldOpt = None
    ) -> None:
        query = SqliteQueryBuilder.build_insert_query(valuelists[0], fields)
        self._executemany(query, valuelists)

    def count_rows(self, clause: str = "") -> int:
        query = f"select count(*) from {{table}} {clause}"
        self._execute(query)
        count = self.controller.fetchone()[0]
        return count

    def get_row(
        self, fields: Iterable = None, clause: str = "", as_dict: bool = False
    ) -> Any:
        fields = iterable_to_fields(fields)
        query = f"select {fields} from {{table}} {clause}"
        self._execute(query)

        row = self.controller.fetchone()
        return sqliterow_to_dict(row) if as_dict else sqliterow_to_tuple(row)

    def get_rows(
        self, fields: Iterable = None, clause: str = "", as_dicts: bool = False
    ) -> list[Any]:
        fields = iterable_to_fields(fields)
        query = f"select {fields} from {{table}} {clause}"
        self._execute(query)

        rows = self.controller.fetchall()
        return sqliterows_to_dicts(rows) if as_dicts else sqliterows_to_tuples(rows)

    def update_rows(self, values: dict, clause: str = None) -> None:
        values_str = ",".join([f"{k} = {v}" for k, v in values.items()])

        clause = clause if clause else str()
        query = f"update {{table}} set {values_str} {clause}"
        self._execute(query)

    def delete_rows(self, clause: str) -> None:
        query = f"delete from {{table}} {clause}"
        self._execute(query)

    def delete_all_rows(self) -> None:
        query = "delete from {table}"
        self._execute(query)


def iterable_to_fields(fields: Optional[Iterable[Any]]) -> str:
    """Convert iterable to query fields string"""
    fields = f"{', '.join(i for i in fields)}" if fields else "*"
    return fields


def sqliterow_to_dict(row):
    """Convert sqlite3.Row instance to dict"""
    return dict(zip(row.keys(), tuple(row)))


def sqliterow_to_tuple(row):
    """Convert sqlite3.Row instance to tuple"""
    return tuple(row)


def sqliterows_to_dicts(rows):
    """Convert sqlite3.Row instances to dicts"""
    for row in rows:
        yield sqliterow_to_dict(row)


def sqliterows_to_tuples(rows):
    """Convert sqlite3.Row instances to tuples"""
    for row in rows:
        yield sqliterow_to_tuple(row)



"""Facilitate handling a SQL database"""

import sqlite3
from abc import ABC, abstractmethod
from typing import Any, Iterable, Optional
from sqlcontroller.validator import AbstractValidator, SqliteValidator
from sqlcontroller.querybuilder import SqliteQueryBuilder
from sqlcontroller.table import DbTable, SqliteTable
from sqlcontroller.field import Field


class NonExistentTableError(Exception):
    """Error for trying to retrieve a non-existent table"""


class AbstractSqlController(ABC):  # pragma: no cover
    """Abstract controller"""

    database: str
    connection: sqlite3.Connection
    cursor: sqlite3.Cursor
    validator: AbstractValidator

    @abstractmethod
    def __enter__(self) -> "AbstractSqlController":
        """Enter controller context"""

    @abstractmethod
    def __exit__(self, *_) -> None:
        """Exit controller context"""

    @abstractmethod
    def connect_db(self) -> sqlite3.Connection:
        """Connect to a database"""

    @abstractmethod
    def disconnect_db(self) -> None:
        """Disconnect from a database"""

    @abstractmethod
    def save_db(self) -> None:
        """Save changes to a database"""

    @abstractmethod
    def get_cursor(self) -> sqlite3.Cursor:
        """Get database cursor"""

    @abstractmethod
    def execute(
        self, query: str, table: str = None, values: Optional[Iterable] = None
    ) -> Any:
        """Execute sql query with one value set"""

    @abstractmethod
    def executemany(
        self,
        query: str,
        table: str = None,
        valuelists: Optional[Iterable[Iterable]] = None,
    ) -> Any:
        """Execute sql query with many value sets"""

    @abstractmethod
    def create_table(self, name: str, fields: Iterable[Field]) -> DbTable:
        """Add new table to a database"""

    @abstractmethod
    def get_table(self, name: str) -> DbTable:
        """Retrieve existing table"""

    @abstractmethod
    def delete_table(self, name: str) -> None:
        """Remove a table from a database"""


class BaseSqlController(AbstractSqlController):
    """Provide generic functionality for an SQL controller"""

    connection: sqlite3.Connection
    cursor: sqlite3.Cursor

    def __init__(self, database):
        self.database = database

    def __enter__(self) -> "BaseSqlController":
        self.connection = self.connect_db()
        self.connection.row_factory = sqlite3.Row
        self.cursor = self.get_cursor()
        return self

    def __exit__(self, *_) -> None:
        self.save_db()
        self.disconnect_db()

    def execute(
        self, query: str, table: str = None, values: Optional[Iterable] = None
    ) -> Any:
        """Execute sql query with one value set"""
        values = values if values else []
        query = query.format(table=table)
        return self.cursor.execute(query, values)

    def executemany(
        self,
        query: str,
        table: str = None,
        valuelists: Optional[Iterable[Iterable]] = None,
    ) -> Any:
        """Execute sql query with many value sets"""
        valuelists = valuelists if valuelists else [[]]
        query = query.format(table=table)
        return self.cursor.executemany(query, valuelists)


class SqliteController(BaseSqlController):
    """Provide methods for database, table and row handling"""

    validator: SqliteValidator = SqliteValidator()

    def connect_db(self) -> sqlite3.Connection:
        """Connect to a database (create if non-existent)"""
        self.connection = sqlite3.connect(self.database)
        return self.connection

    def disconnect_db(self) -> None:
        """Clear database connection"""
        self.connection.close()
        del self.connection
        del self.cursor

    def save_db(self) -> None:
        """Save changes to a database"""
        self.connection.commit()

    def get_cursor(self) -> sqlite3.Cursor:
        """Get database cursor"""
        self.cursor = self.connection.cursor()
        return self.cursor

    def has_table(self, name: str) -> bool:
        """Check if table exists"""
        try:
            self.execute("select * from {table} limit 1", name)
            return True
        except sqlite3.OperationalError:
            return False

    def create_table(self, name: str, fields: Iterable[Field]) -> SqliteTable:
        """Create a new table"""

        self.validator.validate_table_params(name, fields)

        query = SqliteQueryBuilder.build_table_create_query(fields)
        self.execute(query, name)
        return SqliteTable(name, self)

    def get_table(self, name: str) -> SqliteTable:
        """Retrieve existing table"""
        self.validator.validate_table_name(name)

        if not self.has_table(name):
            raise NonExistentTableError(f"{name} table does not exist")

        return SqliteTable(name, self)

    def delete_table(self, name: str) -> None:
        """Delete table"""
        query = "drop table {table};"
        self.execute(query, name)

    @staticmethod
    def build_query_clauses(
        where: str = "", order: str = "", limit: int = 0, offset: int = 0
    ) -> str:
        """Build a query's clauses string"""
        return SqliteQueryBuilder.build_query_clauses(where, order, limit, offset)

    def fetchone(self) -> Any:
        """Get first select query result"""
        return self.cursor.fetchone()

    def fetchall(self) -> list:
        """Get all select query results"""
        return self.cursor.fetchall()



"""SQL validation"""

import re
from abc import ABC
from typing import Any, Iterable
from sqlcontroller.field import Field


class InvalidAlphanumericError(Exception):
    """Error for invalid alphanumeric strings"""


class InvalidSqlDataTypeError(Exception):
    """Error for invalid SQL data types"""


class InvalidSqlDataConstraintError(Exception):
    """Error for invalid SQL data constraints"""


class InvalidSqlOperatorError(Exception):
    """Error for invalid SQL operators"""


class InvalidSqlOperandError(Exception):
    """Error for invalid SQL operands"""


class InvalidSqlNameError(Exception):
    """Error for invalid SQL names"""


class InvalidSqlFieldError(Exception):
    """Error for invalid SQL fields"""


math_comparison = {"=", "<>", "!=", "<", ">", "<=", ">="}
logic_comparison = {"BETWEEN", "EXISTS", "IN", "LIKE"}
types = {"NULL", "INTEGER", "REAL", "TEXT", "BLOB"}
bool_operators = {"ALL", "AND", "ANY", "OR", "NOT"}
constraints = {"NOT NULL", "UNIQUE", "PRIMARY KEY"}


def is_numeric(num: str) -> bool:
    """Return True if string only contains a number"""
    try:
        float(num)
        return True
    except ValueError:
        return False


class AbstractValidity(ABC):  # pylint: disable=too-few-public-methods
    """Abstract class for SQL Validity classes"""


class AbstractValidator(ABC):  # pylint: disable=too-few-public-methods
    """Abstract class for SQL Validator classes"""


class SqliteValidity(AbstractValidity):
    """Check valid SQL keywords, names, types and values"""

    alphanum = re.compile(r"[a-zA-Z0-9]+")
    alphanum_underscore = re.compile(r"[a-zA-Z0-9_]+")
    alphanum_space = re.compile(r"[a-zA-Z0-9\s]+")
    alphanum_underscore_space = re.compile(r"[a-zA-Z0-9\s_]+")

    @staticmethod
    def valid_iterable(var: Any) -> bool:
        """Validate an iterable type"""
        return hasattr(var, "__iter__") and not isinstance(var, str)

    @staticmethod
    def valid_str(var: Any) -> bool:
        """Validate a string type"""
        return isinstance(var, str)

    @staticmethod
    def valid_alphanum(
        name: str, underscore: bool = False, space: bool = False
    ) -> bool:
        """Validate alphanumeric strings (with/without underscores, spaces)"""
        if underscore and space:
            pattern = SqliteValidity.alphanum_underscore_space
        elif underscore:
            pattern = SqliteValidity.alphanum_underscore
        elif space:
            pattern = SqliteValidity.alphanum_space
        else:
            pattern = SqliteValidity.alphanum

        return isinstance(name, str) and bool(re.fullmatch(pattern, name))

    @staticmethod
    def valid_name(name: str) -> bool:
        """Validate sql name"""
        return SqliteValidity.valid_alphanum(name, True, False)

    @staticmethod
    def valid_type(name: str) -> bool:
        """Validate field type"""
        return isinstance(name, str) and name.upper() in types

    @staticmethod
    def valid_constraint(name: str) -> bool:
        """Validate field constraint"""
        return isinstance(name, str) and name.upper() in constraints

    @staticmethod
    def valid_field(field: Field) -> bool:
        """Validate field instance"""
        val = SqliteValidity

        valid_constraints_iterable = lambda: (
            val.valid_iterable(field.constraints)
            and all(list(map(val.valid_constraint, field.constraints)))
        )

        valid_constraints_str = lambda: (
            val.valid_str(field.constraints)
            and val.valid_constraint(field.constraints)  # type: ignore
        )

        valid = (
            isinstance(field, Field)
            and val.valid_name(field.name)
            and val.valid_type(field.type)
            and (valid_constraints_iterable() or valid_constraints_str())
        )

        return valid

    @staticmethod
    def valid_values(values: Iterable) -> bool:
        """Validate values"""

        def is_valid(val: Any) -> bool:
            try:
                return isinstance(val, str) or is_numeric(val)
            except TypeError:
                return False

        return all(map(is_valid, values))

    @staticmethod
    def valid_comparison_operator(comp_op: str) -> bool:
        """Validate comparison operator"""
        return comp_op.upper() in {*math_comparison, *logic_comparison}

    @staticmethod
    def valid_bool_operator(bool_op: str) -> bool:
        """Validate bool operator"""
        return bool_op.upper() in bool_operators


class SqliteValidator(AbstractValidator):
    """Validate SQL keywords, names, types and values"""

    @staticmethod
    def validate_iterable(var: Any) -> None:
        """Validate iterable type"""
        if not SqliteValidity.valid_iterable(var):
            raise TypeError(f"{var} is not iterable")

    @staticmethod
    def validate_str(var: Any) -> None:
        """Validate string type"""
        if not SqliteValidity.valid_str(var):
            raise TypeError(f"{var} is not a string")

    @staticmethod
    def validate_alphanum(
        str_: str, underscore: bool = False, space: bool = False
    ) -> None:
        """Validate alphanumeric string"""
        SqliteValidator.validate_str(str_)

        if not SqliteValidity.valid_alphanum(str_, underscore, space):
            error = f"Not alphanumeric: {str_}"
            raise InvalidAlphanumericError(error)

    @staticmethod
    def validate_table_name(name: str) -> None:
        """Validate field name"""
        if not SqliteValidity.valid_name(name):
            error = f"{name} is not a valid table name."
            raise InvalidSqlNameError(error)

    @staticmethod
    def validate_field_name(field: str) -> None:
        """Validate field name"""
        if not SqliteValidity.valid_name(field):
            error = f"{field} is not a valid field name."
            raise InvalidSqlOperandError(error)

    @staticmethod
    def validate_type(name: str) -> None:
        """Validate field type"""
        SqliteValidator.validate_str(name)

        if not SqliteValidity.valid_type(name):
            error = f"{name} is not a valid field type."
            raise InvalidSqlDataTypeError(error)

    @staticmethod
    def validate_constraint(name: str) -> None:
        """Validate field constraint"""
        SqliteValidator.validate_str(name)

        name = name.upper()

        if not SqliteValidity.valid_constraint(name):
            error = f"{name} is not a valid field constraint."
            raise InvalidSqlDataConstraintError(error)

    @staticmethod
    def validate_field(field: Field) -> None:
        """Validate Field instance"""
        if not SqliteValidity.valid_field(field):
            error = f"{field} is not a valid field."
            raise InvalidSqlFieldError(error)

    @staticmethod
    def validate_table_params(name: str, fields: Iterable[Field]) -> None:
        """Validate table name and fields"""
        SqliteValidator.validate_table_name(name)

        # In case a single field is passed
        if isinstance(fields, Field):
            fields = [fields]

        for field in fields:
            SqliteValidator.validate_field(field)

    @staticmethod
    def validate_values(values: str) -> None:
        """Validate field values"""
        if not SqliteValidity.valid_values(values):
            error = f"{values} is not a valid clause values operand."
            raise InvalidSqlOperandError(error)

    @staticmethod
    def validate_comparison_operator(comp_op: str) -> None:
        """Validate clause comparison operator"""
        if not SqliteValidity.valid_comparison_operator(comp_op):
            error = f"{comp_op} is not a valid comparison operator."
            raise InvalidSqlOperatorError(error)

    @staticmethod
    def validate_bool_operator(bool_op: str) -> None:
        """Validate clause boolean operator"""
        if not SqliteValidity.valid_bool_operator(bool_op):
            error = f"{bool_op} is not a valid boolean operator."
            raise InvalidSqlOperatorError(error)

    @staticmethod
    def validate_condition(field: str, operator: str, values: str) -> None:
        """Validate clause condition"""
        SqliteValidator.validate_field_name(field)
        SqliteValidator.validate_comparison_operator(operator)
        SqliteValidator.validate_values(values)



"""Build SQL queries"""

import re
from abc import ABC
from typing import Collection, Iterable, Optional
from sqlcontroller.field import Field


class BaseSqlQueryBuilder(ABC):  # pylint: disable=too-few-public-methods
    """Abstract query builder"""


class SqliteQueryBuilder(BaseSqlQueryBuilder):  # pylint: disable=too-few-public-methods
    """Build Sqlite query strings"""

    @staticmethod
    def build_table_create_query(fields: Iterable[Field]) -> str:
        """Build a create table query"""

        fields_strs = [" ".join([f.name, f.type, *f.constraints]) for f in fields]
        fields_str = ", ".join(fields_strs)

        query = f"create table if not exists {{table}} ({fields_str});"
        return query

    @staticmethod
    def build_insert_query(
        values: Collection, fields: Optional[Iterable] = None
    ) -> str:
        """Build an insert query string"""

        parts = ["insert into {table}"]

        if fields:
            fields_str = ",".join(fields)
            parts.append(f"({fields_str})")

        parts.append("values")

        qmarks = ",".join(["?"] * len(values))
        parts.append(f"({qmarks})")

        query = " ".join(parts) + ";"
        return query

    @staticmethod
    def build_query_clauses(
        where: str = "", order: str = "", limit: int = 0, offset: int = 0
    ) -> str:
        """Build a query's clauses string"""

        where = re.sub("^where ", "", where, flags=re.IGNORECASE)
        order = re.sub("^order by ", "", order, flags=re.IGNORECASE)

        parts = []
        if where:
            parts.append(f"where {where};")

        if order:
            parts.append(f"order by {order};")

        if limit:
            parts.append(f"limit {limit};")

        if offset:
            parts.append(f"offset {offset};")

        clause = " ".join(parts)
        return clause
